#!/usr/bin/python

import re
from expression import Expression
from tokenizer import Token, Tokenizer

class Parser:
    def __init__(self, query):
        self.__stack = []
        self.__query = query
        self.__tokenizer = Tokenizer(query)
        self.__current = None
        self.__current_lp = None

    def Parse(self):
        has_lp = self.__ParseLeftParen()
        left = self.__ParseIdentifier()
        if self.__Peek().type == Token.Types.KEYWORD:
            expr = self.__ParseMethod(left)
            and_or_expr = self.__ParseAndOr(expr)
            if and_or_expr:
                expr = and_or_expr

            return expr
        else:
            expr = self.__ParseBinary(left)
            if expr:
                if not has_lp or self.__ParseRightParent():
                    and_or_expr = self.__ParseAndOr(expr)
                    if and_or_expr:
                        return and_or_expr

                    return expr

        self.__RaiseError(self.__Peek())

    def debug_string(self):
        r = self.Parse()
        return self.__ToJson(r)

    def __ParseLeftParen(self):
        return self.__ParseDelimiter("(")

    def __ParseRightParent(self):
        return self.__ParseDelimiter(")")

    def __ParseOperator(self):
        t = self.__Peek()
        if t.type == Token.Types.OPERATOR \
                and not re.match("^(and|or)$", self.__Peek().value):
            self.__Accept(Token.Types.OPERATOR)
            return True
        return False

    def __ParseAndOr(self, expr):
        has_lp = self.__ParseLeftParen()
        t = self.__Peek()
        if t.type == Token.Types.OPERATOR \
                and re.match("^(and|or)$", self.__Peek().value):
            self.__Accept(Token.Types.OPERATOR)
            expr = self.__ParseBinary(expr, op=t, right=self.Parse())
            if not has_lp or self.__ParseRightParent():
                return expr

        return None

    def __ParseBinary(self, left, op=None, right=None):
        if not op:
            op = self.__Peek()

        if not right:
            if self.__ParseOperator():
                right = self.__ParseLiteral()
            else:
                right = self.__ParseAndOr(left)

        if right:
            return Expression.New(Expression.BINARY, op.value, left, right)

        return None

    def __ParseIdentifier(self, depth=0):
        t = self.__Peek()
        if self.__Accept(Token.Types.IDENTIFIER):
            if not depth and self.__current_lp and\
                            self.__current_lp.value != t.value:
                self.__RaiseError(t)

            dt = self.__Peek()
            expr = None
            if dt.type == Token.Types.DELIMITER:
                if dt.value == "/":
                    self.__Accept(Token.Types.DELIMITER)
                    it = self.__Peek()
                    if it.type == Token.Types.IDENTIFIER:
                        expr = self.__ParseIdentifier(depth=depth + 1)
            return Expression.New(Expression.MEMBER, t.value, expr)
        return None

    def __ParseDelimiter(self, v):
        lp = self.__Peek()
        if lp.type == Token.Types.DELIMITER and lp.value == v:
            self.__Accept(Token.Types.DELIMITER)
            return True
        return False

    def __ParseLiteral(self):
        l = self.__Peek()
        if l.type == Token.Types.LITERAL:
            self.__Accept(Token.Types.LITERAL)
            return Expression.New(Expression.LITERAL, l.value)

        return None

    def __ParseMethod(self, member):
        k = self.__Peek()
        expr = None
        if k.type == Token.Types.KEYWORD:
            self.__Accept(Token.Types.KEYWORD)
            if self.__ParseLeftParen():
                if k.value == "any":
                    expr = Expression.New(Expression.METHOD_CALL, "any", member, expr=self.__ParseLambda())
                else:
                    expr = Expression.New(Expression.METHOD_CALL, k.value, member, args=self.__ParseMethodArgs())

                if self.__ParseRightParent():
                    return expr

        return None

    def __ParseLambda(self):
        it = self.__Peek()
        if self.__Accept(Token.Types.IDENTIFIER):
            if self.__ParseDelimiter(":"):
                self.__current_lp = it
                pe = Expression.New(Expression.PARAMETER, it.value)
                be = self.Parse()
                self.__current_lp = None
                return Expression.New(Expression.LAMBDA, pe, be)

        return None

    def __ParseMethodArgs(self, list=None):
        if not list:
            list = []

        expr = self.__ParseLiteral()
        if expr:
            list.append(expr)
            if self.__ParseDelimiter(","):
                return self.__ParseMethodArgs(list)

            return list

        self.__RaiseError(self.__Peek())

    def __Accept(self, type):
        if self.__Peek().type == type:
            if self.__tokenizer.HasNext():
                self.__current = self.__tokenizer.GetNextToken()
            return True
        return False

    def __Peek(self):
        if not self.__current:
            if self.__tokenizer.HasNext():
                self.__current = self.__tokenizer.GetNextToken()

        return self.__current

    def __RaiseError(self, *args):
        s = args[0].offset[0]
        e = args[len(args) - 1].offset[1]
        q = self.__query[0:e].split(' ')
        m = q.pop()
        if len(q) > 1:
            m = "%s %s" % (q.pop(), m)
            s = e - len(m)

        raise SyntaxError("Syntax error at position %s-%s on \"%s\"" % (s, e, m))

    def __ToJson(self, r):
        return r.ToJson()



#!/usr/bin/python

import re

class Token:
    class Types:
        IDENTIFIER = "identifier"
        LITERAL = "literal"
        KEYWORD = "keyword"
        OPERATOR = "operator"
        DELIMITER = "delimiter"
        WHITESPACE = "whitespace"

    def __init__(self, value, type, offset):
        self.__value = value
        self.__type = type
        self.__offset = offset

    @property
    def value(self):
        return self.__value

    @property
    def type(self):
        return self.__type

    @property
    def offset(self):
        return self.__offset

    def __eq__(self, other):
        return self.type == other.type \
               and self.value == other.value;

    def __str__(self):
        return "%s, %s | %s | \"%s\"" % (self.__offset[0], self.__offset[1], self.type, self.__value)

class Tokenizer:
    KEYWORD_RE = re.compile("^(any|sub|startswith|endswith|in|nin)$")
    IDENTIFIER_RE = re.compile("^([_a-zA-Z][_a-zA-Z0-9]*)|\$$")
    OPERATOR_RE = re.compile("^(and|eq|(gte?)|(lte?)|ne|or)$")
    DELIMITER_RE = re.compile("[\(\)\/,\:]")
    WHITESPACE_RE = re.compile("^\s+$")
    LITERAL_RE = re.compile("^((\w+)?\'.*\')|(\d+)|(true|false)$")

    def __init__(self, query):
        if not query:
            raise ValueError("query")

        self.__query = query
        self.__pos = 0

    def GetNextToken(self):
        s = ''
        instr = False
        esc = False
        x = None
        sp = self.__pos

        while self.__HasSymbol():
            x = self.__GetNextSymbol()
            if x == ' ':
                if not s:
                    s = x
                    break
                elif not instr:
                    self.__pos -= 1
                    break
            elif re.match(Tokenizer.DELIMITER_RE, x):
                if not instr:
                    if s:
                        self.__pos -= 1
                    else:
                        s = x

                    break
            elif x == '\'':
                if instr:
                    if not esc:
                        s += x
                        break
                else:
                    instr = True
            elif x == '\\':
                if not instr and not esc:
                    esc = True
                    s += x
                    continue
            if esc:
                esc = False

            s += x

        if s:
            t = self.__FindType(s)
            if t == Token.Types.WHITESPACE:
                return self.GetNextToken()

            return Token(s, t, [sp, self.__pos])

        return None

    def HasNext(self):
        return self.__HasSymbol()

    def __HasSymbol(self):
        return self.__pos < len(self.__query)

    def __GetNextSymbol(self):
        c = self.__query[self.__pos]
        self.__pos = self.__pos + 1
        return c

    def __FindType(self, value):
        if not value:
            raise ValueError("value")

        if self.__IsWhitespace(value):
            return Token.Types.WHITESPACE
        elif self.__IsLiteral(value):
            return Token.Types.LITERAL
        elif self.__IsDelimiter(value):
            return Token.Types.DELIMITER
        elif self.__IsKeyword(value):
            return Token.Types.KEYWORD
        elif self.__IsOperator(value):
            return Token.Types.OPERATOR
        elif self.__IsIdentifier(value):
            return Token.Types.IDENTIFIER

        return None

    def __IsWhitespace(self, value):
        return re.match(Tokenizer.WHITESPACE_RE, value)

    def __IsDelimiter(self, value):
        return re.match(Tokenizer.DELIMITER_RE, value)

    def __IsKeyword(self, value):
        return re.match(Tokenizer.KEYWORD_RE, value)

    def __IsOperator(self, value):
        return re.match(Tokenizer.OPERATOR_RE, value)

    def __IsIdentifier(self, value):
        return re.match(Tokenizer.IDENTIFIER_RE, value)

    def __IsLiteral(self, value):
        return re.match(Tokenizer.LITERAL_RE, value)


#!/usr/bin/python

class Expression(object):
    BINARY = "binary"
    MEMBER = "member"
    PARAMETER = "parameter"
    LITERAL = "literal"
    METHOD_CALL = "method_call"
    LAMBDA = "lambda"

    def __init__(self, type):
        self._type = type

    @staticmethod
    def New(type, *args, **kwargs):
        if type == Expression.MEMBER:
            return MemberExpression(*args, **kwargs)

        if type == Expression.PARAMETER:
            return ParameterExpression(*args)

        if type == Expression.LITERAL:
            return LiteralExpression(*args)

        if type == Expression.BINARY:
            return BinaryExpression(*args)

        if type == Expression.METHOD_CALL:
            return MethodCallExpression(*args, **kwargs)

        if type == Expression.LAMBDA:
            return LambdaExpression(*args)

        raise ValueError("invalid expression!")

    @property
    def type(self):
        return self._type

class BinaryExpression(Expression):
    def __init__(self, operator, left, right):
        super(BinaryExpression, self) \
            .__init__(Expression.BINARY)

        self.__operator = operator
        self.__left = left
        self.__right = right

    @property
    def right(self):
        return self.__right

    @property
    def operator(self):
        return self.__operator

    @property
    def left(self):
        return self.__left

    def ToJson(self):
        return dict(
            type=self.type,
            operator=self.__operator,
            left=self.__left.ToJson(),
            right=self.__right.ToJson()
        )

class MemberExpression(Expression):
    def __init__(self, name, expr=None):
        if not name:
            raise AttributeError("name")

        self.__name = name
        self.__expr = expr

        super(MemberExpression, self) \
            .__init__(Expression.MEMBER)

    @property
    def name(self):
        return self.__name

    @property
    def expr(self):
        return self.__expr

    def ToJson(self):
        return dict(
            type=self.type,
            name=self.__name,
            expr=self.__expr.ToJson() if self.__expr else None
        )

class ParameterExpression(Expression):
    def __init__(self, name):
        if not name:
            raise AttributeError("name")

        self.__parameter = name
        super(ParameterExpression, self) \
            .__init__(Expression.PARAMETER)

    @property
    def parameter(self):
        return self.__parameter

    def ToJson(self):
        return dict(
            type=self.type,
            parameter=self.parameter
        )

class LiteralExpression(Expression):
    def __init__(self, value):
        self.__value = value
        super(LiteralExpression, self) \
            .__init__(Expression.LITERAL)

    @property
    def value(self):
        return self.__value

    def ToJson(self):
        return dict(
            type=self.type,
            value=self.__value
        )

class MethodCallExpression(Expression):
    def __init__(self, name, member, expr=None, args=None):
        self.__name = name
        self.__member = member
        self.__args = args
        self.__expr = expr

        super(MethodCallExpression, self) \
            .__init__(Expression.METHOD_CALL)

    @property
    def name(self):
        return self.__name

    @property
    def member(self):
        return self.__member

    @property
    def args(self):
        return self.__args

    @property
    def expr(self):
        return self.__expr

    def ToJson(self):
        args = []
        if self.__args:
            for x in self.__args:
                args.append(x.ToJson())

        return dict(
            type=self.type,
            name=self.__name,
            member=self.__member.ToJson(),
            expr=self.__expr.ToJson() if self.__expr else None,
            args=args
        )

class LambdaExpression(Expression):
    def __init__(self, parameter, body):
        if not parameter:
            raise AttributeError("parameter")

        self.__body = body
        self.__parameter = parameter

        super(LambdaExpression, self) \
            .__init__(Expression.LAMBDA)

    @property
    def parameter(self):
        return self.__parameter

    @property
    def body(self):
        return self.__body

    def ToJson(self):
        return dict(
            type=self.type,
            parameter=self.__parameter.ToJson(),
            body=self.__body.ToJson()
        )

class ExpressionVisitor(object):
    def Visit(self, expr):
        if expr.type == Expression.BINARY:
            return self.VisitBinary(expr)
        elif expr.type == Expression.MEMBER:
            return self.VisitMember(expr);
        elif expr.type == Expression.PARAMETER:
            return self.VisitParameter(expr);
        elif expr.type == Expression.LITERAL:
            return self.VisitLiteral(expr);
        elif expr.type == Expression.METHOD_CALL:
            return self.VisitMethodCall(expr);
        elif expr.type == Expression.LAMBDA:
            return self.VisitLambda(expr);

    def VisitBinary(self, expr):
        pass

    def VisitMember(self, expr):
        pass

    def VisitParameter(self, expr):
        pass

    def VisitLiteral(self, expr):
        pass

    def VisitMethodCall(self, expr):
        pass

    def VisitLambda(self, expr):
        pass


#!/usr/bin/python

from distutils.core import setup

setup(
    name='catq',
    version='0.2.0',
    packages=['catq'],
    description='Simple resource query language.',
    author='Ebubekir Dirican',
    author_email='ebubekirdirican@gmail.com',
    url='https://github.com/ebbkrdrcn/catq'
)